variogram <- function(data,dt=NULL,fast=NULL,CI="Markov")
{
  if(length(dt)<2) { return(variogram.dt(data,dt=dt,fast=fast,CI=CI)) }
  
  var <- lapply(dt, function(DT) { variogram.dt(data,dt=DT,fast=fast,CI=CI) } )
  
  dt <- c(dt,Inf)

  class(var[[1]]) <- "data.frame"
  var[[1]] <- subset(var[[1]],lag<dt[2])
  for(i in 2:(length(dt)-1))
  {
    class(var[[i]]) <- "data.frame"
    var[[i]] <- subset(var[[i]],lag<dt[i+1])[-1,]
  }
  
  # lazy coalate
  var <- mean.variogram(var)
  
  return(var)
} 
  

# wrapper for fast and slow variogram codes, for a specified dt
variogram.dt <- function(data,dt=NULL,fast=NULL,CI="Markov")
{
  # intelligently select algorithm
  if(is.null(fast))
  {
    if(length(data$t)<1000) { fast <- FALSE }
    else { fast <- TRUE }
  }
  
  if(fast)
  {
    return(variogram.fast(data=data,dt=dt,CI=CI))
  }
  else
  { 
    return(variogram.slow(data=data,dt=dt,CI=CI))
  }
}

############################
# FFT VARIOGRAM
# not to be exposed to end users
variogram.fast <- function(data,dt=NULL,CI="Markov")
{
  # quick replacement for fftw
  FFT <- function(X) { fft(X,inverse=FALSE) }
  IFFT <- function(X) { fft(X,inverse=TRUE)/length(X) }
  
  t <- data$t
  x <- data$x
  y <- data$y
  
  n <- length(t)
  
  # time lags
  DT <- diff(t)
 
  # default time step
  if(is.null(dt))
  { dt <- median(DT) }
  
  # gap weights to prevent oversampling with coarse dt
  W <- clamp(c(DT[1],DT)/dt) # left weights
  W <- W + clamp(c(DT,DT[n-1])/dt) # right weights
  W <- W/2 # average left and right
  
  # choose best grid alignment
  t <- t - t[1]
  cost <- function(t0)
  { # not sure if t and W get inherited
    grid <- t-t0
    return( sum(W*(grid-round(grid))^2) )
  }
  t <- t - optimize(cost,c(-dt,dt)/2)$minimum
  
  # count times from first grid point
  t <- t - floor(t[1]/dt)*dt
  
  # fractional grid index
  index <- 1 + t/dt
  
  # uniform lag grid
  lag <- seq(0,ceiling(t[n]/dt)-floor(t[1]/dt))*dt
  n <- length(lag)
  
  # continuously distribute times over uniform grid
  W.grid <- rep(0,n)
  X.grid <- rep(0,n)
  Y.grid <- rep(0,n)
  for(i in 1:length(t))
  {
    j <- index[i]
    
    if(floor(j)==ceiling(j))
    { # trivial case
      J <- round(j)
      w <- W[i] # total weight
      W.grid[J] <- W.grid[J] + w
      X.grid[J] <- X.grid[J] + w*x[i]
      Y.grid[J] <- Y.grid[J] + w*y[i]
    }
    else
    { # distribution information between adjacent grids
     
      # left grid portion
      J <- floor(j)
      w <- W[i]*(1-(j-J))
      W.grid[J] <- W.grid[J] + w
      X.grid[J] <- X.grid[J] + w*x[i]
      Y.grid[J] <- Y.grid[J] + w*y[i]
   
      # right grid portion
      J <- ceiling(j)
      w <- W[i]*(1-(J-j))
      W.grid[J] <- W.grid[J] + w
      X.grid[J] <- X.grid[J] + w*x[i]
      Y.grid[J] <- Y.grid[J] + w*y[i]
    }
  }
  
  # normalize distributed information
  for(i in 1:n)
  {
    if(W.grid[i]>0)
    {
      X.grid[i] <- X.grid[i]/W.grid[i]
      Y.grid[i] <- Y.grid[i]/W.grid[i]
    }
  }
  # continuous weights fuck up the FFT numerics. discretize weights for now.
  W <- sum(W) # now total DOF
  W.grid <- sign(W.grid) # discrete weights
  
  W.grid <- Conj(FFT(pad(W.grid,2*n)))
  XX.grid <- FFT(pad(X.grid^2,2*n))
  YY.grid <- FFT(pad(Y.grid^2,2*n))
  X.grid <- FFT(pad(X.grid,2*n))
  Y.grid <- FFT(pad(Y.grid,2*n))

  # pair number. one for x and y data
  DOF <- round(Re(2*IFFT(abs(W.grid)^2)[1:n]))
  # SVF un-normalized
  SVF <- Re(IFFT(Re(W.grid*(XX.grid+YY.grid))-(abs(X.grid)^2+abs(Y.grid)^2))[1:n])
  
  # delete missing lags
  SVF <- data.frame(SVF=SVF,DOF=DOF,lag=lag)
  SVF <- subset(SVF,DOF>0)
  lag <- SVF$lag
  DOF <- SVF$DOF
  SVF <- SVF$SVF
  
  # normalize SVF
  SVF <- SVF/DOF
  
  # only count non-overlapping lags... not perfect
  if(CI=="Markov")
  {
    dof <- 2*(t[length(t)]-t[1])/lag
    dof[1] <- 2*length(t)
  
    for(i in 1:length(lag))
    {
      if(dof[i]<DOF[i]) {DOF[i] <- dof[i] }
    }
  }
  else if(CI=="IID") # fix initial and total DOF
  {
    DOF[1] <- 2*length(t)
    DOF[-1] <- DOF[-1]/sum(DOF[-1])*(length(t)^2-length(t))
  }
  
  result <- data.frame(SVF=SVF,DOF=DOF,lag=lag)
  class(result) <- "variogram"
  
  return(result)
}

##################################
# LAG-WEIGHTED VARIOGRAM
# not to be exposed to end users
variogram.slow <- function(data,dt=NULL,CI="Markov")
{
  t <- data$t
  x <- data$x
  y <- data$y

  n <- length(t)
  
  # time lags
  DT <- diff(t)
  DT.L <- c(DT[1],DT)
  DT.R <- c(DT,DT[n-1])
  
  # default time step
  if(is.null(dt))
  { dt <- median(DT) }

  # where we will store stuff
  lag <- seq(0,ceiling((t[n]-t[1])/dt))*dt
  SVF <- rep(0,length(lag))
  DOF <- rep(0,length(lag))
  DOF2 <- rep(0,length(lag))
  
  pb <- txtProgressBar(style=3)
  for(i in 1:n)
  { 
    for(j in i:n)
    {
      tau <- t[j] - t[i]
      var <- ((x[j]-x[i])^2 + (y[j]-y[i])^2)/4
      
      # gap weight
      if(tau==0) { w <- 1 }
      else { w <- (clamp(DT.L[j]/tau)+clamp(DT.R[j]/tau))*(clamp(DT.L[i]/tau)+clamp(DT.R[i]/tau)) }
      
      # fractional index
      k <- tau/dt + 1
      
      if(floor(k)==ceiling(k))
      { # even sampling
        # lag index
        K <- round(k)
        # total weight
        W <- w
        # accumulate
        SVF[K] <- SVF[K] + W*var
        DOF[K] <- DOF[K] + W
        DOF2[K] <- DOF2[K] + W^2
      }
      else
      { # account for drift by distributing semi-variance
        
        # left index
        K <- floor(k)
        # left weight
        W <- w*(1-(k-K))
        # accumulate left portion
        SVF[K] <- SVF[K] + W*var
        DOF[K] <- DOF[K] + W
        DOF2[K] <- DOF2[K] + W^2
        
        # right index
        K <- ceiling(k)
        # right weight
        W <- w*(1-(K-k))
        # accumulate right portion
        SVF[K] <- SVF[K] + W*var
        DOF[K] <- DOF[K] + W
        DOF2[K] <- DOF2[K] + W^2
      }
    }
    setTxtProgressBar(pb,(i*(2*n-i))/(n^2))
  }
  
  # delete missing lags
  SVF <- data.frame(SVF=SVF,DOF=DOF,DOF2=DOF2,lag=lag)
  SVF <- subset(SVF,DOF>0)
  lag <- SVF$lag
  DOF <- SVF$DOF
  DOF2 <- SVF$DOF2
  SVF <- SVF$SVF
  
  # normalize SVF
  SVF <- SVF/DOF
  # effective DOF from weights, one for x and y
  DOF <- 2*DOF^2/DOF2
  
  # only count non-overlapping lags... still not perfect
  if(CI=="Markov")
  {
    dof <- 2*length(t)
    if(dof<DOF[1]) { DOF[1] <- dof  }
    
    for(i in 2:length(lag))
    { # large gaps are missing data
      dof <- 2*sum(DT[DT<=lag[i]])/lag[i]
      if(dof<DOF[i]) { DOF[i] <- dof }
      
      setTxtProgressBar(pb,i/length(lag))
    }
  }
  else if(CI=="IID") # fix initial and total DOF
  {
    DOF[1] <- 2*length(t)
    DOF[-1] <- DOF[-1]/sum(DOF[-1])*(length(t)^2-length(t))
  }
  
  close(pb)
  
  result <- data.frame(SVF=SVF,DOF=DOF,lag=lag)
  class(result) <- "variogram"
  
  return(result)
}



subset.variogram <- function(variogram,...)
{
  class(variogram) <- "data.frame"
  variogram <- subset(variogram,...)
  class(variogram) <- "variogram"
  return(variogram)
}



###########################################################
# PLOT VARIOGRAM
###########################################################
plot.variogram <- function(variogram, model=NULL, alpha=0.05, fraction=0.5, col.variogram="red", col.model="blue", ...)
{  
  # number of variograms
  if(class(variogram)=="variogram") { variogram <- list(variogram) }
  n <- length(variogram)
  
  # maximum lag in data
  max.lag <- sapply(variogram, function(v){ last(v$lag) } )
  max.lag <- max(max.lag)
  # subset fraction of data
  max.lag <- fraction*max.lag
  
  # subset all data to fraction of total period
  for(i in 1:n) { class(variogram[[i]]) <- "data.frame" }
  variogram <- lapply(variogram, function(v) { subset(v, lag <= max.lag) })

  # confidence interval functions
  CI.upper <- Vectorize(function(k,Alpha){qchisq(Alpha/2,k,lower.tail=FALSE)/k})
  CI.lower <- Vectorize(function(k,Alpha){qchisq(Alpha/2,k,lower.tail=TRUE)/k})
  
  # maximum CI on SVF
  max.SVF <- max(sapply(variogram, function(v){ max(v$SVF * CI.upper(v$DOF,min(alpha))) } ))
  # limit plot range to twice max SVF point estimate (otherwise hard to see)
  max.cap <- 2*max(sapply(variogram, function(v){ max(v$SVF) } ))
  if(max.SVF>max.cap) { max.SVF <- max.cap }
  
  # choose SVF units
  SVF.scale <- unit(max.SVF,"area")
  SVF.name <- SVF.scale$name
  SVF.scale <- SVF.scale$scale
  
  # choose lag units
  lag.scale <- unit(max.lag,"time",2)
  lag.name <- lag.scale$name
  lag.scale <- lag.scale$scale
  
  xlab <- paste("Time-lag ", "(", lag.name, ")", sep="")
  ylab <- paste("Semi-variance ", "(", SVF.name, ")", sep="")
  
  # fix base plot layer
  plot(c(0,max.lag/lag.scale),c(0,max.SVF/SVF.scale), xlab=xlab, ylab=ylab, col=rgb(1,1,1,0), ...)
  
  # color array for plots
  col <- array(col.variogram,n)
  
  for(i in 1:n)
  {
    lag <- variogram[[i]]$lag/lag.scale
    SVF <- variogram[[i]]$SVF/SVF.scale
    DOF <- variogram[[i]]$DOF
        
    # make sure plot looks nice and appropriate for data resolution
    type <- "l"
    if(length(lag) < 100) { type <- "p" }
    
    points(lag, SVF, type=type, col=col[[i]])
    
    for(j in 1:length(alpha))
    {
      SVF.lower <- SVF * CI.lower(DOF,alpha[j])
      SVF.upper <- SVF * CI.upper(DOF,alpha[j])
      
      polygon(c(lag,rev(lag)),c(SVF.lower,rev(SVF.upper)),col=translucent(col[[i]],alpha=0.1/length(alpha)),border=NA)
    }
  }
  
  # NOW PLOT THE MODELS
  if(!is.null(model))
  {
    if(class(model)=="ctmm") { model <- list(model) }
    n <- length(model) 
    
    # color array for plots
    col <- array(col.model,n)
    type <- "l"
        
    for(i in 1:n)
    {
      # pull out relevant model parameters
      tau <- model[[i]]$tau/lag.scale
      sigma <- model[[i]]$sigma/SVF.scale
      COV <- model[[i]]$COV.tau
      
      tau <- tau[tau>0]
      K <- length(tau)
      
      # trace variance
      if(length(sigma)==1) { sigma <- sigma * diag(2) } # for compatibility
      GM.sigma <- sqrt(det(sigma))
      sigma <- mean(diag(sigma)) # now AM.sigma
      
      # parameter covariances
      if(is.null(COV)) { COV <- diag(0,K+1) } # default to no error considered
      
      # unit conversion for plot
      vec <- c(rep(1/lag.scale,K),1/SVF.scale)
      COV <- diag(vec,nrow=K+1) %*% COV %*% diag(vec,nrow=K+1)
      
      # transform to a simpler parameterization
      f <- 1/tau
      vec <- c(f^2,1)
      COV <- diag(vec,nrow=K+1) %*% COV %*% diag(vec,nrow=K+1)
      
      K <- length(tau)
      if(K==0)
      { 
        svf <- Vectorize( function(t){ if(t==0) {0} else {sigma} } )
        d.svf <- function(t){ if(t==0) {0} else {sigma/GM.sigma} }
      }
      else if(K==1)
      {
        svf <- Vectorize( function(t){ sigma*(1-exp(-t*f)) } )
        d.svf <- function(t){ c( sigma*(-t*exp(-t*f)) , svf(t)/GM.sigma ) }
      }
      else if(K==2)
      { 
        svf <- Vectorize( function(t){ sigma*(1 - (f[2]*exp(-t*f[1])-f[1]*exp(-t*f[2]))/(f[2]-f[1])) } )
        d.svf <- function(t)
        { c(
          -sigma*(-t*f[2]*exp(-t*f[1])-exp(-t*f[2]))/(f[2]-f[1]) - sigma*(f[2]*exp(-t*f[1])-f[1]*exp(-t*f[2]))/(f[2]-f[1])^2,
          -sigma*(exp(-t*f[1])+t*f[1]*exp(-t*f[2]))/(f[2]-f[1]) + sigma*(f[2]*exp(-t*f[1])-f[1]*exp(-t*f[2]))/(f[2]-f[1])^2,
          svf(t)/GM.sigma
        ) }
      }

      # point estimate
      curve(svf, from=0, to=lag[length(lag)], add=TRUE, type=type, col=col[[i]])
      
      # confidence intervals if COV provided
      if(any(diag(COV)>0))
      {
        Lags <- seq(lag[1],last(lag),(last(lag)-lag[1])/1000)
        
        # efffective DOF of SVF if its chi square
        DOF.svf <- Vectorize(function(t){ 2*svf(t)^2/( d.svf(t) %*% COV %*% d.svf(t) ) })
        
        for(j in 1:length(alpha))
        {
          svf.lower <- function(t){ svf(t) * CI.lower(DOF.svf(t),alpha[j]) }
          svf.upper <- function(t){ svf(t) * CI.upper(DOF.svf(t),alpha[j]) }
          
          polygon(c(Lags,rev(Lags)),c(svf.lower(Lags),rev(svf.upper(Lags))),col=translucent(col[[i]],alpha=0.1/length(alpha)),border=NA)
        }
      }
    }
  }
  
}


#######################################
# plot a variogram with zoom slider
#######################################
variogram.zoom <- function(variogram,log=TRUE,...)
{
  # number of variograms
  n <- 1
  if(class(variogram)=="list") { n <- length(variogram) }
  else {variogram <- list(variogram) } # promote to list of one
  
  # maximum lag in data
  max.lag <- sapply(variogram, function(v){ last(v$lag) } )
  max.lag <- max(max.lag)
  
  min.lag <- sapply(variogram, function(v){ v$lag[2] } )
  min.lag <- min(min.lag)
  
  min.step <- min.lag/max.lag
  
  if(!log)
  {
    manipulate::manipulate(
    { plot.variogram(variogram,fraction=fraction,...) },
      fraction = manipulate::slider(10*min.step, 1.0,initial=0.5,step=min.step)
    )
  }
  else
  { 
    b <- 10
    manipulate::manipulate(
    { plot.variogram(variogram,fraction=min.step*b^(z),...) },
      z = manipulate::slider(log(10,b),log(1/min.step,b),initial=log(0.5/min.step,b),label="zoom")
    )
  }
  
}

######################################################################
# visual fit of the variogram
######################################################################
variogram.fit <- function(variogram,sigma.max=NULL,tau.max=NULL,log=TRUE,...)
{
  K <- length(tau.max)
  m <- 3 # slider length
  
  # variance slider parameters
  sigma <- mean(variogram$SVF)
  if(is.null(sigma.max)) { sigma.max <- m*sigma }
  
  if(K<2)
  {
    # peak curvature estimate
    v2 <- 2*max((variogram$SVF/variogram$lag^2)[2:length(variogram$lag)])
    # peak diffusion rate estimate
    D <- max((variogram$SVF/variogram$lag)[2:length(variogram$lag)])
    # velocity timescale estimate 
    tau <- D/v2
    tau.max <- m*tau
    
    # position timescale estimate
    tau <- c(tau,sigma/D)
    tau.max <- c(tau.max,m*tau[2])
  }
  else
  {
   tau.int <- tau.max/2 
  }
  
  min.step <- variogram$lag[2]/last(variogram$lag)
    
  # I can't figure out how to store by argument/pointer
  
  if(log)
  {
    b <- 10
    manipulate::manipulate(
      {
      if(store) { variogram.fit.model <<- list(sigma=sigma,tau=c(tau.v,tau.r)) ; class(variogram.fit.model) <<- "ctmm" }
      model <- list(sigma=sigma,tau=c(tau.v,tau.r))
      class(model) <- "ctmm"
      plot.variogram(variogram,model=model,fraction=min.step*b^(z),...)
      },
      z = manipulate::slider(log(10,b),log(1/min.step,b),initial=log(0.5/min.step,b),label="zoom"),
      sigma = manipulate::slider(0,sigma.max,initial=sigma,label="sigma (m^2)"),
      tau.r = manipulate::slider(0,tau.max[2],initial=tau[2],label="tau[2] (s)"),
      tau.v = manipulate::slider(0,tau.max[1],initial=tau[1],label="tau[1] (s)"),
      store = manipulate::button("save to variogram.fit.model")
    )
  }
  else
  {
    manipulate::manipulate(
    {
    if(store) { variogram.fit.model <<- list(sigma=sigma,tau=c(tau.v,tau.r)) ; class(variogram.fit.model) <<- "ctmm" }
    model <- list(sigma=sigma,tau=c(tau.v,tau.r))
    class(model) <- "ctmm"
    plot.variogram(variogram,model=model,fraction=fraction,...)
    },
    fraction = manipulate::slider(0.0, 1.0,initial=0.5,step=min.step),
    sigma = manipulate::slider(0,sigma.max,initial=sigma,label="sigma (m^2)"),
    tau.r = manipulate::slider(0,tau.max[2],initial=tau[2],label="tau[2] (s)"),
    tau.v = manipulate::slider(0,tau.max[1],initial=tau[1],label="tau[1] (s)"),
    store = manipulate::button("save to variogram.fit.model")
    )
  }
  
#  model <<- list(sigma=sigma,tau=tau) ; class(model) <<- "ctmm"
#  return(model)
  
}


# AVERAGE VARIOGRAMS
mean.variogram <- function(varlist)
{
  IDS <- length(varlist)
  
  # assemble observed lag range
  lag.min <- rep(0,IDS) # this will be dt
  lag.max <- rep(0,IDS)
  for(id in 1:IDS)
  { 
    n <- length(varlist[[id]]$lag)
    lag.max[id] <- varlist[[id]]$lag[n]
    lag.min[id] <- varlist[[id]]$lag[2]
  }
  lag.max <- max(lag.max)
  lag.min <- min(lag.min)
  lag <- seq(0,ceiling(lag.max/lag.min))*lag.min
  
  # where we will store everything
  n <- length(lag)
  SVF <- rep(0,n)
  DOF <- rep(0,n)
  
  # accumulate semivariance
  for(id in 1:IDS)
  {
   for(i in 1:length(varlist[[id]]$lag))
   {
    # lag index
    j <- 1 + round(varlist[[id]]$lag[i]/lag.min)
    # number weighted accumulation
    DOF[j] <- DOF[j] + varlist[[id]]$DOF[i]
    SVF[j] <- SVF[j] + varlist[[id]]$DOF[i]*varlist[[id]]$SVF[i]
   }
  }
  
  # delete missing lags
  variogram <- data.frame(SVF=SVF,DOF=DOF,lag=lag)
  variogram <- subset(variogram,DOF>0)

  # normalize SVF
  variogram$SVF <- variogram$SVF / variogram$DOF
  
  class(variogram) <- "variogram"
  
  return(variogram)
}
